#
# Copyright (c) 2021-2022 LAAS/CNRS
# All rights reserved.
#
# Redistribution  and  use  in  source  and binary  forms,  with  or  without
# modification, are permitted provided that the following conditions are met:
#
#   1. Redistributions of  source  code must retain the  above copyright
#      notice and this list of conditions.
#   2. Redistributions in binary form must reproduce the above copyright
#      notice and  this list of  conditions in the  documentation and/or
#      other materials provided with the distribution.
#
# THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
# WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
# MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
# ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
# WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
# IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#
#                                           Anthony Mallet on Tue Dec 14 2021
#
"""Low level interaction with a genomix server."""

import warnings

from . import component
from . import event
from . import http
from . import request


# --- connect --------------------------------------------------------------
#
def connect(host = 'localhost:8080'):
  """Connect to remote genomixd server and return the connection handle."""
  return Handle(host)


class Handle:
  """Active connection to a genomix server."""

  # --- __init__ -----------------------------------------------------------
  #
  def __init__(self, host, session = ''):
    """Open a new connection with the genomixd server on host."""
    self.host = host
    self.__pending = request.Handles()
    self.__session = None

    # connect to host
    self.__http = http.Client(host)

    # get server version
    r = self.__get('/version')
    try:
      v = tuple(map(int, (str(r).split("."))))
    except:
      raise RuntimeError('not a genomixd server') from None
    if v < (1, 7, 1):
      raise RuntimeError(
        'unsupported genomixd server version %s, need >= 1.7.1' % r)

    # create session
    self.__session = self.__get('/session/start', {'s': session})

    # create SSE connection
    self.__sse = http.Client(host)
    r = self.__sse.get('/session/listen', {'s': self.__session})
    if not 200 <= r['status'] < 300:
      raise Exception(r['body'])
    self.__sse.watch(self.__event)


  # --- __repr__ -----------------------------------------------------------
  #
  def __repr__(self):
    return '<genomix connection at %s%s>' % (
      self.host,
      ' [closed]' if self.__session is None else ''
    )


  # --- close --------------------------------------------------------------
  #
  def close(self):
    """Terminate a connection."""

    # flush any pending events
    event.mux.wait(0)

    # close sockets
    self.__sse.close()
    self.__http.close()
    self.__session = None

    # mark components unloaded
    self.__pending.close(...)
    component.Handle.close(self, ...)


  # --- read ---------------------------------------------------------------
  #
  def reconnect(self):
    """Reopen a connection with the genomixd server."""

    self.close()
    self.__init__(self.host)


  # --- read ---------------------------------------------------------------
  #
  def read(self, client, port):
    """Read the given port of the component."""
    return self.__get('/module/%s/read/%s' % (client, port))


  # --- send ---------------------------------------------------------------
  #
  def send(self, client, service, args):
    """Send a request for the given service of the component."""
    rid = str(self.__get('/module/%s/send/%s' % (client, service), args))
    if not rid: return None

    return self.__pending(self, client, rid)


  # --- rpath --------------------------------------------------------------
  #
  def rpath(self, path):
    """Configure remote load path."""
    return self.__get('/rpath/%s' % path)


  # --- load ---------------------------------------------------------------
  #
  def load(self, client, *args):
    """Create a component client instance."""
    return component.load(self, client, *args)


  # --- path ---------------------------------------------------------------
  #
  def path(self, component):
    """Return the path where a component client was loaded from."""
    return self.__get('/module/%s/path' % component)


  # --- unload -------------------------------------------------------------
  #
  def unload(self, component):
    """Unload a component client."""
    return self.__get('/module/%s/unload' % component)


  # --- options ------------------------------------------------------------
  #
  def options(self, client):
    """Return load options dictionary of a client."""
    return self.__get('/usage/%s' % client)


  # --- info ---------------------------------------------------------------
  #
  def info(self, component):
    """Return meta information for a component."""
    return self.__get('/module/%s/info' % component)


  # --- dlopen -------------------------------------------------------------
  #
  def dlopen(self, client, options):
    """Load a component client in the genomixd server."""
    return self.__get('/load/%s' % client, {'argv': ' '.join(options)})


  # --- abort --------------------------------------------------------------
  #
  def abort(self, component, rid):
    """Request for cancellation of the request in the component."""
    r = self.__pending(None, component, rid)
    if not r: raise RuntimeError('bad request id')

    self.__get('/module/%s/abort/%s' % (component, rid))


  # --- clean --------------------------------------------------------------
  #
  def clean(self, component, rid):
    """Discard a request id so that no further events will be reported."""

    self.__pending.clean(component, rid)
    self.__get('/module/%s/clean/%s' % (component, rid))


  # --- __get --------------------------------------------------------------
  #
  def __get(self, uri, args = {}, body = None):
    """Make an HTTP get transaction."""
    if ('s' not in args and self.__session is not None):
      args['s'] = self.__session

    r = self.__http.get(uri, args, body)
    body = r['body']

    if not 200 <= r['status'] < 300:
      if isinstance(body, dict): raise event.GenoMError(body)
      else: raise RuntimeError(body)

    return body


  # --- __event ------------------------------------------------------------
  #
  def __event(self, data = None):
    for e in self.__sse.events(data):
      if 'request' in e['event']:
        self.__pending.reply(e['data'])
      elif 'unload' in e['event']:
        self.__pending.close(e['data'])
        component.Handle.close(self, e['data'])
      elif 'quit' in e['event']:
        self.close()
      else:
        warnings.warn(RuntimeWarning('unknown %s event' % e['event']))
